#include "stdafx.h"

#include "Dependency.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

// PATTERN: Double Dispatch

class FrameIntersectsTarget : public std::unary_function<PFrameOperation, bool>,
	public IOperationVisitor
{
private:
	LONG m_frame;
	mutable bool m_result;
protected:
	void setResult(bool result) const
	{
		m_result = result;
	}
public:
	explicit FrameIntersectsTarget(LONG frame)
		: m_frame(frame)
	{
	}

	bool operator()(const PFrameOperation& operation) const
	{
		if (operation->targetFrame() != m_frame) return false;
		m_result = true;
		operation.accept(*const_cast<FrameIntersectsTarget*>(this));
		return m_result;
	}

	void operator()(FramePatch&) {}
	void operator()(FieldPatch&) {}
	void operator()(FrameImage&) {}
	void operator()(FieldRebuild&) {}
	void operator()(MotionInterpolate&) {}
};

class FieldIntersectsTarget : public FrameIntersectsTarget
{
private:
	LONG m_order;
public:
	FieldIntersectsTarget(LONG frame, LONG order)
		: FrameIntersectsTarget(frame), m_order(order)
	{
	}

	using FrameIntersectsTarget::operator();

	void operator()(FieldPatch& lhs)
	{
		setResult(lhs.targetOrder() == m_order);
	}
};

class FrameRectIntersectsTarget : public FrameIntersectsTarget
{
private:
	CRect m_rect;
public:
	FrameRectIntersectsTarget(LONG frame, const CRect& rect)
		: FrameIntersectsTarget(frame), m_rect(rect)
	{
		m_rect.NormalizeRect();
	}

	using FrameIntersectsTarget::operator();

	void operator()(FramePatch& lhs)
	{
		CRect otherRect = lhs.targetRect();
		otherRect.NormalizeRect();
		setResult(!(m_rect & otherRect).IsRectEmpty());
	}

	void operator()(FieldPatch& lhs)
	{
		CRect otherRect = lhs.targetRect();
		otherRect.NormalizeRect();
		setResult(!(m_rect & otherRect).IsRectEmpty());
	}
};

class FieldRectIntersectsTarget : public FrameRectIntersectsTarget
{
private:
	LONG m_order;
public:
	FieldRectIntersectsTarget(LONG frame, LONG order, const CRect& rect)
		: FrameRectIntersectsTarget(frame, rect), m_order(order)
	{
	}

	using FrameRectIntersectsTarget::operator();

	void operator()(FieldPatch& lhs)
	{
		if (m_order != lhs.targetOrder())
			return setResult(false);
		FrameRectIntersectsTarget::operator()(lhs);
	}
};

class TrueDependency : public std::unary_function<PFrameOperation, bool>,
	public IOperationVisitor
{
private:
	PFrameOperation m_lhs;
	mutable bool m_result;
public:
	TrueDependency(const PFrameOperation& lhs)
		: m_lhs(lhs)
	{
	}

	bool operator()(const PFrameOperation& rhs) const
	{
		rhs.accept(*const_cast<TrueDependency*>(this));
		return m_result;
	}

	void operator()(FramePatch& rhs)
	{
		m_result = FrameRectIntersectsTarget(rhs.sourceFrame(), rhs.sourceRect())(m_lhs);
	}

	void operator()(FieldPatch& rhs)
	{
		m_result = FieldRectIntersectsTarget(rhs.sourceFrame(), rhs.sourceOrder(), rhs.sourceRect())(m_lhs);
	}

	void operator()(FrameImage& rhs)
	{
		m_result = false;
	}

	void operator()(FieldRebuild& rhs)
	{
		m_result = FieldIntersectsTarget(rhs.targetFrame() - 1, 0)(m_lhs) ||
			FieldIntersectsTarget(rhs.targetFrame() + 1, 1)(m_lhs);
	}

	void operator()(MotionInterpolate& rhs)
	{
		m_result = FrameIntersectsTarget(rhs.targetFrame() - 1)(m_lhs) ||
			FrameIntersectsTarget(rhs.targetFrame() + 1)(m_lhs);
	}
};

class OutputDependency : public std::unary_function<PFrameOperation, bool>,
	public IOperationVisitor
{
private:
	PFrameOperation m_lhs;
	mutable bool m_result;
public:
	OutputDependency(const PFrameOperation& lhs)
		: m_lhs(lhs)
	{
	}

	bool operator()(const PFrameOperation& rhs) const
	{
		rhs.accept(*const_cast<OutputDependency*>(this));
		return m_result;
	}

	void operator()(FramePatch& rhs)
	{
		m_result = FrameRectIntersectsTarget(rhs.targetFrame(), rhs.targetRect())(m_lhs);
	}

	void operator()(FieldPatch& rhs)
	{
		m_result = FieldRectIntersectsTarget(rhs.targetFrame(), rhs.targetOrder(), rhs.targetRect())(m_lhs);
	}

	void operator()(FrameImage& rhs)
	{
		m_result = FrameIntersectsTarget(rhs.targetFrame())(m_lhs);
	}

	void operator()(FieldRebuild& rhs)
	{
		m_result = FrameIntersectsTarget(rhs.targetFrame())(m_lhs);
	}

	void operator()(MotionInterpolate& rhs)
	{
		m_result = FrameIntersectsTarget(rhs.targetFrame())(m_lhs);
	}
};

bool DependencyExists::operator()(const PFrameOperation& lhs, const PFrameOperation& rhs) const
{
	return        lhs && rhs && (
		TrueDependency(lhs)(rhs) ||
		TrueDependency(rhs)(lhs) ||
		OutputDependency(lhs)(rhs));
}
