#pragma once

#include <windows.h>
#include <vfw.h>

#include <atlimage.h>

#include <vector>
#include <map>

#include "refcount.h"
#include "exception.h"

LONG align(LONG n, LONG alignment);

class PFrame;

class Bitmap : public ref_counted
{
private: // noncopyable
	Bitmap(const Bitmap& other);
	Bitmap& operator=(const Bitmap& other);

private:
	//HBITMAP m_handle;
	CImage m_image;
public:
	HBITMAP handle() const;

private:
	~Bitmap() {}

public:
	CSize size() const;
	LONG width() const;
	LONG height() const;
	CRect rect() const;

public:
	explicit Bitmap(LPCSTR lpszFileName);
};

typedef counted_ptr<Bitmap> PBitmap;

class Frame : public ref_counted
{
private:
	BITMAPINFOHEADER m_header;
	RGBQUAD m_colors[256];

	void Frame::assignHeader(const BITMAPINFOHEADER& header);
	BITMAPINFO* bitmapInfo() { return reinterpret_cast<BITMAPINFO*>(&m_header); }

	//HBITMAP m_handle;
	//VOID* m_pData;

	HANDLE m_hMapping;
	VOID* m_pData;

	void alloc(DWORD size);

	Frame& operator=(const Frame& other); // noncopyable

	~Frame();
public:
	explicit Frame(const BITMAPINFOHEADER& header, int width, int height);
	explicit Frame(LPVOID pData);
	explicit Frame(const BITMAPINFOHEADER& header, const Bitmap& bitmap);
	Frame(const Frame& other);

	HANDLE hMapping() const { return m_hMapping; }
	LPBITMAPINFOHEADER header() const;
	LPVOID data() const;

	void dump() const;
};

// PFrame is an intrusive reference-counting smart pointer/reference
// to a Frame

class SeparateFields
{
public:
	void operator()(BITMAPINFOHEADER& bih) const;
};

class PFrame : public counted_ptr<Frame>
{
private:
	typedef counted_ptr<Frame> Base;
	LPVOID m_pData;
public:
	LPVOID data() const;

	const RGBTRIPLE& at(LONG x, LONG y) const;
	RGBTRIPLE& at(LONG x, LONG y);

	template <typename TransformT>
	void copyFrom(const PFrame& source, CRect sourceRect, CRect targetRect, const TransformT& transform);
	void copyFrom(const PFrame& source, CRect sourceRect, CRect targetRect);
	void copyFrom(const PBitmap& source);

private:
	BITMAPINFOHEADER m_header;
	RGBQUAD m_colors[256];
	void assignHeader(const BITMAPINFOHEADER* pHeader);
public:
	const BITMAPINFOHEADER* header() const;
	LPBITMAPINFOHEADER header();
	const BITMAPINFO* bitmapInfo() const;
	LONG width() const;
	LONG height() const;
	CSize size() const;
	CRect rect() const;
	SIZE_T stride() const;

	CPoint fieldOrigin(LONG order) const;

private:
	HBITMAP m_handle;
public:
	HBITMAP handle() const { return m_handle; }

private:
	int PFrame::lineFromY(int y) const;
	SIZE_T PFrame::byteFromX(int x) const;
	LPVOID PFrame::ptr(int x, int y) const;

public:
	PFrame(Frame* frame = 0);
	PFrame(const PFrame& other);

	void swap(PFrame& other);
	PFrame& operator=(const PFrame& other);

	~PFrame();

	PFrame clone() const;
	
	template <typename HeaderTransformT>
	PFrame(const PFrame& other, const HeaderTransformT& transform)
	: Base(other), m_header(other.m_header), m_handle(0), m_pData(0)
	{
		transform(m_header);
		if (*this)
		{
			m_handle = ASSERT_NE(CreateDIBSection(0, bitmapInfo(), DIB_RGB_COLORS,
				&m_pData, get()->hMapping(), 0), HBITMAP(), "CreateDIBSection");
		}
	}
};

// AviFile is a wrapper for AVIFile, AVIStream and GetFrame,
// with a cache of frames.
class AviFile
{
private:
	AviFile(const AviFile& other); // noncopyable
	AviFile& operator=(const AviFile& other); // noncopyable

	PAVIFILE m_pAviFile;
	PAVISTREAM m_pAviStream;
	PGETFRAME m_pGetFrame;
	LONG m_length;
	LONG m_millifps;
	LONG m_width;
	LONG m_height;

	typedef std::map<LONG, PFrame> Frames;
	mutable Frames m_frames;
public:
	AviFile();
	~AviFile();

	void open(LPCTSTR lpszFilename);
	void close();

	PFrame operator[](LONG frame) const;
	PFrame getUncached(LONG frame) const;
	void erase(LONG frame);
	LONG size() const;
	LONG millifps() const;
	LONG width() const;
	LONG height() const;
	CRect frameSize() const;
};

class ImageCache
{
private: // noncopyable
	ImageCache(const ImageCache& other);
	ImageCache& operator=(const ImageCache& other);

private:
	typedef std::map<std::string, PBitmap> Images;
	mutable Images m_images;

public:
	ImageCache() {}
	PBitmap operator[](const std::string& fileName) const;
};

#include "FrameCache.inline.h"
