#ifndef EXCEPTION_INCLUDED
#define EXCEPTION_INCLUDED

#include <exception>
#include <stdexcept>
#include <string>
#include <cstring>
#include <sstream>

#include <windows.h>
#include <gdiplus.h>

class TraceContext
{
private:
	static std::ostringstream s_trace;
public:
	static const std::string trace()
	{
		return s_trace.str();
	}
	static void clear()
	{
		s_trace.str("");
	}

private:
	std::ostringstream m_trace;
public:
	template <typename Ostreamable>
	TraceContext& operator()(const Ostreamable& value)
	{
		m_trace << value;
		return *this;
	}

	~TraceContext()
	{
		if (std::uncaught_exception())
		{
			s_trace << m_trace.str() << "\n";
		}
	}
};

#define YK_TRACED(x) TraceContext yk_trace_context; yk_trace_context(x)

class Win32Exception : public std::runtime_error
{
public:
	Win32Exception(const std::string& what_arg)
	: std::runtime_error(what_arg), code_(GetLastError())
	{
		OutputDebugString(what());
	}

	Win32Exception(const std::string& what_arg, HRESULT hResult)
	: std::runtime_error(what_arg), code_(hResult)
	{
		OutputDebugString(what());
	}

	DWORD code() const
	{
		return code_;
	}

	virtual const char* what() const throw()
	{
		static char message[4096];
		char* end = message + sizeof(message);
		std::strncpy(message, std::runtime_error::what(), sizeof(message) - 1);
		*(end - 1) = 0;

		char* cur = message + std::strlen(message);
		*cur++ = ':';
		*cur++ = ' ';

		if (!FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM, 0, code_, 0, cur, DWORD(end - cur), 0))
		{
			_snprintf(cur, end - cur - 1, "Error code %d", code_);
			*(end - 1) = 0;
		}
		return message;
	}
private:
	DWORD code_;
};

class GdiplusException : public std::runtime_error
{
private:
	Gdiplus::Status status_;
public:
	GdiplusException(const std::string& what_arg, Gdiplus::Status status)
		: std::runtime_error(what_arg), status_(status)
	{
		OutputDebugString(what());
	}

	Gdiplus::Status status() const
	{
		return status_;
	}

	virtual const char* what() const throw()
	{
		static char message[4096];
		char* end = message + sizeof(message);
		std::strncpy(message, std::runtime_error::what(), sizeof(message) - 1);
		*(end - 1) = 0;

		char* cur = message + std::strlen(message);
		*cur++ = ':';
		*cur++ = ' ';
		size_t remain = sizeof(message) - (cur - message);

		switch (status_)
		{
#define YK_GDIP_STATUS(status) \
		case Gdiplus::status: \
			strncpy(cur, #status, remain); \
			break;
			YK_GDIP_STATUS(Ok)
			YK_GDIP_STATUS(GenericError)
			YK_GDIP_STATUS(InvalidParameter)
			YK_GDIP_STATUS(OutOfMemory)
			YK_GDIP_STATUS(ObjectBusy)
			YK_GDIP_STATUS(InsufficientBuffer)
			YK_GDIP_STATUS(NotImplemented)
			YK_GDIP_STATUS(Win32Error)
			YK_GDIP_STATUS(WrongState)
			YK_GDIP_STATUS(Aborted)
			YK_GDIP_STATUS(FileNotFound)
			YK_GDIP_STATUS(ValueOverflow)
			YK_GDIP_STATUS(AccessDenied)
			YK_GDIP_STATUS(UnknownImageFormat)
			YK_GDIP_STATUS(FontFamilyNotFound)
			YK_GDIP_STATUS(FontStyleNotFound)
			YK_GDIP_STATUS(NotTrueTypeFont)
			YK_GDIP_STATUS(UnsupportedGdiplusVersion)
			YK_GDIP_STATUS(GdiplusNotInitialized)
			YK_GDIP_STATUS(PropertyNotFound)
			YK_GDIP_STATUS(PropertyNotSupported)
#undef YK_GDIP_STATUS
		default:
			itoa(status_, cur, 10);
		}

		message[sizeof(message) - 1] = 0;

		return message;
	}
};

template <typename Value, typename ErrorValue>
Value assertNotEqual(Value result, ErrorValue errorValue, const char* location, const char* message)
{
	if (result == errorValue)
	{
		throw Win32Exception(std::string(location) + message);
	}
	return result;
}

inline HRESULT assertSuccess(HRESULT hResult, const char* location, const char* message)
{
	if (FAILED(hResult))
	{
		throw Win32Exception(std::string(location) + message, hResult);
	}
	return hResult;
}

inline int assertFar(int result, const char* location, const char* message)
{
	if (!result)
	{
		throw std::runtime_error(std::string(location) + message);
	}
	return result;
}

inline LONG assertRegistry(LONG result, const char* location, const char* message)
{
	if (ERROR_SUCCESS != result)
	{
		throw Win32Exception(std::string(location) + message, result);
	}
	return result;
}

inline LONG assertGdip(Gdiplus::Status result, const char* location, const char* message)
{
	if (Gdiplus::Ok != result)
	{
		throw GdiplusException(std::string(location) + message, result);
	}
	return result;
}

#define STRINGIZE1(x) #x
#define STRINGIZE(x) STRINGIZE1(x)

//#ifdef DEBUG

#define ASSERT_NE(result, errorValue, message) assertNotEqual(result, errorValue, __FILE__ ":" STRINGIZE(__LINE__) ": ", message)
#define ASSERT_SUCCESS(hResult, message) assertSuccess(hResult, __FILE__ ":" STRINGIZE(__LINE__) ": ", message)
#define ASSERT_FAR(result, message) assertFar(result, __FILE__ ":" STRINGIZE(__LINE__) ": ", message)
#define ASSERT_REGISTRY(result, message) assertRegistry(result, __FILE__ ":" STRINGIZE(__LINE__) ": ", message)
#define ASSERT_GDIP(result, message) assertGdip(result, __FILE__ ":" STRINGIZE(__LINE__) ": ", message)

//#else
//
//#define ASSERT_NE(result, errorValue, message) (result)
//#define ASSERT_SUCCESS(hResult, message) (hResult)
//#define ASSERT_FAR(result, message) (result)
//#define ASSERT_REGISTRY(result, message) (result)
//
//#endif

#endif
